# from ..nonlinear import *
# import numpy as np
# from nose import tools
# 
# class TestNonlinearSimulation(object):
#     
#     wildtype = "00"
#     mutations = {0:["0","1"], 1:["0","1"]}
# 
#     def test_init(self):
#         sim = NonlinearSimulation(self.wildtype, self.mutations)
#         # Checks
#         assert isinstance(sim, NonlinearSimulation)
